
#include "RobboLito.h"
#include <string.h>

#define MAX_AGE 256
#define MAX_DEPTH 256

#define AGE_DEPTH_MIX(x,y)  ((( (x) - AGE) & (MAX_AGE - 1)) * MAX_DEPTH   + (MAX_DEPTH - (y) ))

void
IncrementAge ()
{
  AGE += 1;
  if (AGE == MAX_AGE)
    AGE = 0;
}

static uint64 HASH_SIZE = 0x400000;
static boolean FLAG_HASH_INIT = 0;

void
HashClear ()
{
  memset (HashTable, 0, HASH_SIZE * sizeof (typeHash));
  memset (PVHashTable, 0, 0x10000 * sizeof (typePVHash));
  AGE = 0;
}

int
InitHash (int mb)
{
  AGE = 0;
  HASH_SIZE = ((1ULL << BSR (mb)) << 20) / sizeof (typeHash);
  if (HASH_SIZE > 0x100000000)
    HASH_SIZE = 0x100000000;
  mb = (HASH_SIZE * sizeof (typeHash)) >> 20;
  HashMask = HASH_SIZE - 4;
  if (FLAG_HASH_INIT)
    free (HashTable);
  FLAG_HASH_INIT = TRUE;
  POSIX_MEMALIGN (HashTable, 64, HASH_SIZE * sizeof (typeHash));
  HashClear ();
  return mb;
}

void
HashLowerALL (typePOS *POSITION, int move, int depth, int Value)
{
  int DEPTH, i, k = DYN->HASH & HashMask;
  typeHash *tr;
  int max = 0, w = 0;
  move &= 0x7fff;
  for (i = 0; i < 4; i++)
    {
      tr = HashTable + (k + i);
      if ((tr->hash ^ (DYN->HASH >> 32)) == 0
	  && (!tr->DepthLower || IsALL (tr))
	  && tr->DepthLower <= depth)
	{
	  tr->DepthLower = depth;
	  tr->move = move;
	  tr->ValueUpper = Value;
	  tr->age = AGE;
	  tr->flags |= FLAG_LOWER | FLAG_ALL;
	  return;
	}
      DEPTH = MAX (tr->DepthLower, tr->DepthUpper);
      if (AGE_DEPTH_MIX (tr->age, DEPTH) > max)
	{
	  max = AGE_DEPTH_MIX (tr->age, DEPTH);
	  w = i;
	}
    }
  tr = HashTable + (k + w);
  tr->hash = (DYN->HASH >> 32);
  tr->DepthUpper = 0;
  tr->ValueLower = 0;
  tr->DepthLower = depth;
  tr->move = move;
  tr->ValueUpper = Value;
  tr->age = AGE;
  tr->flags = FLAG_LOWER | FLAG_ALL;
  return;
}

void
HashUpperCUT (typePOS *POSITION, int depth, int Value)
{
  int DEPTH, i, k = DYN->HASH & HashMask;
  typeHash *tr;
  int max = 0, w = 0;
  for (i = 0; i < 4; i++)
    {
      tr = HashTable + (k + i);
      if (!(tr->hash ^ (DYN->HASH >> 32))
	  && (!tr->DepthUpper || IsCUT (tr))
	  && tr->DepthUpper <= depth)
	{
	  tr->DepthUpper = depth;
	  tr->ValueLower = Value;
	  tr->age = AGE;
	  tr->flags |= FLAG_UPPER | FLAG_CUT;
	  return;
	}
      DEPTH = MAX (tr->DepthLower, tr->DepthUpper);
      if (AGE_DEPTH_MIX (tr->age, DEPTH) > max)
	{
	  max = AGE_DEPTH_MIX (tr->age, DEPTH);
	  w = i;
	}
    }
  tr = HashTable + (k + w);
  tr->hash = (DYN->HASH >> 32);
  tr->DepthLower = 0;
  tr->move = 0;
  tr->ValueUpper = 0;
  tr->DepthUpper = depth;
  tr->ValueLower = Value;
  tr->age = AGE;
  tr->flags = FLAG_UPPER | FLAG_CUT;
  return;
}

void
HashLower (uint64 Z, int move, int depth, int Value)
{
  int DEPTH, i, k = Z & HashMask;
  typeHash *tr;
  int max = 0, w = 0;
  move &= 0x7fff;
  for (i = 0; i < 4; i++)
    {
      tr = HashTable + (k + i);
      if (!(tr->hash ^ (Z >> 32)) && !IsExact (tr)
	  && tr->DepthLower <= depth)
	{
	  tr->DepthLower = depth;
	  tr->move = move;
	  tr->ValueUpper = Value;
	  tr->age = AGE;
	  tr->flags |= FLAG_LOWER;
	  tr->flags &= ~FLAG_ALL;
	  return;
	}
      DEPTH = MAX (tr->DepthLower, tr->DepthUpper);
      if (AGE_DEPTH_MIX (tr->age, DEPTH) > max)
	{
	  max = AGE_DEPTH_MIX (tr->age, DEPTH);
	  w = i;
	}
    }
  tr = HashTable + (k + w);
  tr->hash = (Z >> 32);
  tr->DepthUpper = 0;
  tr->ValueLower = 0;
  tr->DepthLower = depth;
  tr->move = move;
  tr->ValueUpper = Value;
  tr->age = AGE;
  tr->flags = FLAG_LOWER;
  return;
}

void
HashUpper (uint64 Z, int depth, int Value)
{
  int DEPTH, i, k = Z & HashMask;
  typeHash *tr;
  int max = 0, w = 0;
  for (i = 0; i < 4; i++)
    {
      tr = HashTable + (k + i);
      if (!(tr->hash ^ (Z >> 32)) && !IsExact (tr)
	  && tr->DepthUpper <= depth)
	{
	  tr->DepthUpper = depth;
	  tr->ValueLower = Value;
	  tr->age = AGE;
	  tr->flags |= FLAG_UPPER;
	  tr->flags &= ~FLAG_CUT;
	  return;
	}
      DEPTH = MAX (tr->DepthLower, tr->DepthUpper);
      if (AGE_DEPTH_MIX (tr->age, DEPTH) > max)
	{
	  max = AGE_DEPTH_MIX (tr->age, DEPTH);
	  w = i;
	}
    }
  tr = HashTable + (k + w);
  tr->hash = (Z >> 32);
  tr->DepthLower = 0;
  tr->move = 0;
  tr->ValueUpper = 0;
  tr->DepthUpper = depth;
  tr->ValueLower = Value;
  tr->age = AGE;
  tr->flags = FLAG_UPPER;
  return;
}

static void
pv_zobrist (typePOS *POSITION, int move, int depth, int Value)
{
  int i, k = DYN->HASH & PVHashMask;
  typePVHash *tr;
  int w = 0, max = 0;
  for (i = 0; i < 4; i++)
    {
      tr = PVHashTable + (k + i);
      if (tr->hash == DYN->HASH)
	{
	  tr->depth = depth; /* esta seguro que depth es mejor ? */
	  tr->Value = Value;
	  tr->move = move;
	  tr->age = AGE;
	  return;
	}
      if (AGE_DEPTH_MIX (tr->age, tr->depth) > max)
	{
	  max = AGE_DEPTH_MIX (tr->age, tr->depth);
	  w = i;
	}
    }
  tr = PVHashTable + (k + w);
  tr->hash = DYN->HASH;
  tr->depth = depth;
  tr->move = move;
  tr->Value = Value;
  tr->age = AGE;
}

void
HashExact
 (typePOS *POSITION, int move, int depth, int Value, int FLAGS)
{
  int DEPTH, i, j, k = DYN->HASH & HashMask;
  typeHash *tr;
  int max = 0, w = 0;
  move &= 0x7fff;
  pv_zobrist (POSITION, move, depth, Value);
  for (i = 0; i < 4; i++)
    {
      tr = HashTable + (k + i);
      if ((tr->hash ^ (DYN->HASH >> 32)) == 0
	  && MAX (tr->DepthUpper, tr->DepthLower) <= depth)
	{
	  tr->DepthUpper = tr->DepthLower = depth;
	  tr->move = move;
	  tr->ValueLower = tr->ValueUpper = Value;
	  tr->age = AGE;
	  tr->flags = FLAGS;
	  for (j = i + 1; j < 4; j++)
	    {
	      tr = HashTable + (k + j);
	      if ((tr->hash ^ (DYN->HASH >> 32)) == 0
		  && MAX (tr->DepthUpper, tr->DepthLower) <= depth)
		{
		  memset (tr, 0, 16);
		  tr->age = AGE ^ (MAX_AGE / 2);
		}
	    }
	  return;
	}
      DEPTH = MAX (tr->DepthLower, tr->DepthUpper);
      if (AGE_DEPTH_MIX (tr->age, DEPTH) > max)
	{
	  max = AGE_DEPTH_MIX (tr->age, DEPTH);
	  w = i;
	}
    }
  tr = HashTable + (k + w);
  tr->hash = (DYN->HASH >> 32);
  tr->DepthUpper = tr->DepthLower = depth;
  tr->move = move;
  tr->ValueLower = tr->ValueUpper = Value;
  tr->age = AGE;
  tr->flags = FLAGS;
  return;
}
