
#include "RobboLito.h"
#include <string.h>
#include "control.h"

static sint64 LAST_MESSAGE;
static sint64 ABSOLUTE_TIME, DESIRED_TIME, INCREMENT;
static int DEPTH;
static boolean NEW_PONDERHIT;

#if 0
#define ALTERNATIVE_TIME_USAGE TRUE /* en UCI ? */
#endif

void ponderhit() /* ABSOLUTE_TIME se cambia ? */
{
  if (!DO_PONDER)
    return;
  PONDER_HIT = TRUE;
  DO_PONDER = FALSE;
  NEW_PONDERHIT = TRUE;
#if 1 /* ? */
  ABSOLUTE_TIME += GetClock() - START_CLOCK;
#endif
}

void
HaltSearch (int tr)
{
  STOP = TRUE;
  if (JUMP_IS_SET)
    longjmp (J, 1);
}

void
Info (sint64 x)
{
  uint64 t, nps, NODES = 0;
  int cpu, rp;
  clock_t u;
  if (TITANIC_MODE)
    for (cpu = 0; cpu < CPUS_SIZE; cpu++)
      for (rp = 0; rp < RP_PER_CPU ; rp++)
        NODES += ROOT_POSITION[cpu][rp].nodes;
  else
    NODES = ROOT_POSIZIONE0->nodes;
  u = clock ();
  LAST_MESSAGE = x;
  t = x / 1000;
  if (t == 0)
    nps = 0;
  else
    nps = NODES / t;
  u = clock () - INIT_CLOCK;
  SEND ("info time %lld nodes %lld nps %lld cpuload %d\n",
		t, NODES, nps * 1000,
		(int) ((double) u / (double) x * 1000.0));
}

static sint64 BATTLE_TIME, EASY_TIME, ORDINARY_TIME;

#ifdef ALTERNATIVE_TIME_USAGE
int mossa_en_kontrol, ora_en_totale_bianco, ora_en_totale_nero,old_mtg;
boolean INCREASED_DEPTH;
float time_factor;
#endif

void
CheckDone (typePOS *POSITION, int d)
{
  sint64 x;
  if (!JUMP_IS_SET)
    return;
  x = GetClock () - START_CLOCK;
  if (d == DEPTH)
    HaltSearch (1);
  if (x - LAST_MESSAGE > 1000000)
    Info (x);
#if 1
  if (DO_PONDER)
    goto END; /* Decembrist */
  if (DO_INFINITE)
    goto END; /* Decembrist */
#endif
  if (d >= 1 && d < 8)
    goto END;
  if (x > ABSOLUTE_TIME)
    HaltSearch (1);
  if (d == 0 && !NEW_PONDERHIT)
    goto END;
  NEW_PONDERHIT = FALSE;
#if 0
  printf("%lld %lld %lld %lld %x %x %x %x\n",
	 x, ABSOLUTE_TIME, ORDINARY_TIME, BATTLE_TIME,
	 NEW_PONDER, d, BAD_MOVE, BATTLE_MOVE);
#endif
#ifdef ALTERNATIVE_TIME_USAGE
#define INCREASE_DEPTH() { INCREASED_DEPTH = TRUE; goto END; }
#define CHECK_INCREASE_DEPTH()
{ if (INCREASED_DEPTH && time_factor > 1.05) INCREASE_DEPTH (); HaltSearch (2); }
  if (!BAD_MOVE && x >= BATTLE_TIME)
    CHECK_INCREASE_DEPTH();
  if (EASY_MOVE && x >= EASY_TIME)
    CHECK_INCREASE_DEPTH();
  if (!BATTLE_MOVE && x >= ORDINARY_TIME && !BAD_MOVE)
    CHECK_INCREASE_DEPTH();
#else
  if (!BAD_MOVE && x >= BATTLE_TIME)
    HaltSearch (2);
  if (EASY_MOVE && x >= EASY_TIME)
    HaltSearch (3);
  if (!BATTLE_MOVE && x >= ORDINARY_TIME && !BAD_MOVE)
    HaltSearch (4);
#endif
 END:
  while (TryInput ())
    Input (POSITION);
}

#define INFINITY 0x7ffffffffffffff /* sint64 */
#define STRTOK(p) p = strtok (NULL, " ")

static uint32
FullMove (typePOS *POSITION, uint32 x)
{
  int pi, to = TO (x), fr = FROM (x);
  if (!x)
    return x;
  pi = POSITION->sq[fr];
  if (pi == wEnumK || pi == bEnumK)
    {
      if (to - fr == 2 || fr - to == 2)
        x |= FlagOO;
    }
  if (TO (x) != 0 && TO (x) == DYN->ep
      && (pi == wEnumP || pi == bEnumP))
    x |= FlagRP;
  return x;
}

static uint32 numeric_move (typePOS* POSITION, char* str)
{
  int x;
  x = FullMove (POSITION, (str[2] - 'a') + ((str[3] - '1') << 3) +
		      ((str[0] - 'a') << 6) + ((str[1] - '1') << 9));
  if (str[4] == 'b')
    x |= FlagPromB;
  if (str[4] == 'n')
    x |= FlagPromN;
  if (str[4] == 'r')
    x |= FlagPromR;
  if (str[4] == 'q')
    x |= FlagPromQ;
  return x;
}

#ifdef ALTERNATIVE_TIME_USAGE
void
time_manager_009 (uint64 TOTAL_TIME, uint64 TIME, uint64 INCREMENT, int mtg)
{
  TIME = MAX (TIME - 500000, (9 * TIME) / 10);
  if (mtg)
    {
      old_mtg = mtg;
      time_factor = (float) (TIME * mossa_en_kontrol) / (float) (TOTAL_TIME * mtg);
      if (time_factor < 1.0)
	DESIRED_TIME = TIME / mtg + INCREMENT;
      else
	DESIRED_TIME =
	  MIN (TIME * time_factor * time_factor / mtg + INCREMENT, TIME);
      if (time_factor < 1.2 && mtg > 2)
	ABSOLUTE_TIME =
	  MIN ((TIME * time_factor * (mtg + 2)) / (mtg + 3),
		  6 * DESIRED_TIME);
      else
	ABSOLUTE_TIME = (TIME * (mtg + 1)) / (mtg + 2);
      ABSOLUTE_TIME = MAX (10000, ABSOLUTE_TIME);
    }
#define NUMBER 4 /* or 6 ? */
  else /* no mtg */
    {
      int temp = TIME / 20;
      float sf;
      if (temp > INCREMENT)
	{
	  DESIRED_TIME =
	    INCREMENT + (temp * TIME) / ((50 - NUMBER) * temp - 20 * INCREMENT);
	  ABSOLUTE_TIME = (temp * TIME) / (5 * temp - 3 * INCREMENT);
	}
      else
	{
	  ABSOLUTE_TIME = TIME >> 1;
	  ABSOLUTE_TIME = INCREMENT + TIME / (30 - NUMBER);
	}
      sf = (float) (TIME + 20 * INCREMENT) / (float) (TOTAL_TIME);
      if (6.0 * sf < 1.0)
	ABSOLUTE_TIME /= 2;
      else if (3.0 * sf < 1.0)
	ABSOLUTE_TIME *= (3.0 * sf);
      if (TIME < 500000)
	{
	  ABSOLUTE_TIME = 10000;
	  DESIRED_TIME = 5000;
	}
      if (TIME >= 500000 && TIME < 1000000 && INCREMENT < 500000)
	{
	  ABSOLUTE_TIME = (TIME * 5) / 100;
	  DESIRED_TIME = (TIME * 5) / 400;
	}
      if (TIME >= 1000000 && TIME < 2000000 && INCREMENT < 500000)
	{
	  ABSOLUTE_TIME = TIME / 10;
	}

    }
  EASY_TIME = DESIRED_TIME >> 2;
  BATTLE_TIME = DESIRED_TIME;
  ORDINARY_TIME = (3 * DESIRED_TIME) >> 2;
}
#endif

void time_manager (uint64 TIME, uint64 INCREMENT, int mtg)
{  
  if (mtg)
    {
      if (mtg > 25)
	mtg = 25;
      DESIRED_TIME = TIME / mtg + INCREMENT;
      ABSOLUTE_TIME =
	(TIME * mtg) / (2 * mtg - 1) - MIN (1000000, TIME / 10);
      if (ABSOLUTE_TIME < 5000)
	ABSOLUTE_TIME = 5000;
    }
  else
    {
      ABSOLUTE_TIME = TIME / 2 - 10000;
      if (ABSOLUTE_TIME < 5000)
	ABSOLUTE_TIME = 5000;
      DESIRED_TIME = ABSOLUTE_TIME / 22 + INCREMENT;
    }
  if (DESIRED_TIME > ABSOLUTE_TIME)
    DESIRED_TIME = ABSOLUTE_TIME;
  if (DESIRED_TIME < 5000)
    DESIRED_TIME = 5000;

#if 1
  EASY_TIME = DESIRED_TIME / 4;
  BATTLE_TIME = DESIRED_TIME;
  ORDINARY_TIME = (3 * DESIRED_TIME) / 4;
#else
  EASY_TIME = DESIRED_TIME / 4;
  BATTLE_TIME = DESIRED_TIME;
  ORDINARY_TIME = (1 * DESIRED_TIME) / 2;
#endif
}

void
InitSearch (typePOS *POSITION, char *str)
{
  char *p;
  sint64 wtime = INFINITY, btime = INFINITY,
    winc = 0, binc = 0, TIME, mtg = 0;
  int sm = 0;

  DEPTH = 255;
  ABSOLUTE_TIME = DESIRED_TIME = INFINITY;
  STOP = FALSE;
  DO_INFINITE = FALSE;
  DO_PONDER = FALSE;
  NEW_PONDERHIT = FALSE;
  DO_SEARCH_MOVES = FALSE;
  LAST_MESSAGE = 0;
  p = strtok (str, " ");
  for (STRTOK (p); p != NULL; STRTOK (p))
    {
      if (!strcmp (p, "depth"))
	{
	  STRTOK (p);
	  DEPTH = MAX (1, atoi (p));
	}
      else if (!strcmp (p, "movetime"))
	{
	  STRTOK (p);
	  ABSOLUTE_TIME = MAX (1, atoll (p)) * 1000; /* - 10000 ? */
	}
      else if (!strcmp (p, "wtime"))
	{
	  STRTOK (p);
	  wtime = atoll (p) * 1000;
	}
      else if (!strcmp (p, "winc"))
	{
	  STRTOK (p);
	  winc = atoll (p) * 1000;
	}
      else if (!strcmp (p, "btime"))
	{
	  STRTOK (p);
	  btime = atoll (p) * 1000;
	}
      else if (!strcmp (p, "binc"))
	{
	  STRTOK (p);
	  binc = atoll (p) * 1000;
	}
      else if (!strcmp (p, "movestogo"))
	{
	  STRTOK (p);
	  mtg = atoi (p);
	}
      else if (!strcmp (p, "infinite"))
	DO_INFINITE =  TRUE;
      else if (!strcmp (p, "ponder"))
	DO_PONDER = TRUE;
      else if (!strcmp (p, "searchmoves"))
	DO_SEARCH_MOVES = TRUE;
      else if (DO_SEARCH_MOVES)
	SEARCH_MOVES[sm++] = numeric_move (POSITION, p);
      else
	MYERROR ("go string: %s", p);
    }
  BATTLE_TIME = INFINITY;
  ORDINARY_TIME = INFINITY;
  EASY_TIME = INFINITY;
#ifdef ALTERNATIVE_TIME_USAGE
  time_factor = 1.0;
  INCREASED_DEPTH = FALSE;
  if (NEW_GAME || mtg > old_mtg) /* reset */
    {
      mossa_en_kontrol = mtg;
      ora_en_totale_bianco = MAX (wtime - 500000, (95 * wtime) / 100);
      ora_en_totale_nero = MAX (wtime - 500000, (95 * btime) / 100);
    }
#endif
  TIME = POSITION->wtm ? wtime : btime;
  if (TIME == INFINITY)
    goto END;
  INCREMENT =
    POSITION->wtm ? winc : binc;
#ifdef ALTERNATIVE_TIME_USAGE
  time_manager_009
    (POSITION->wtm ?
     ora_en_totale_bianco : ora_en_totale_nero,
     TIME, INCREMENT, mtg); /* alternativo */
#else
  time_manager (TIME, INCREMENT, mtg);
#endif
 END:
  if (TIME == INFINITY)
    ANALYSING = TRUE;
  else
    ANALYSING = FALSE;
  if (DO_SEARCH_MOVES)
    SEARCH_MOVES[sm] = MOVE_NONE;
}
